/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.communication.bus
 * File:    Processor.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.communication.bus;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import de.fhhannover.inform.iron.mapserver.utils.NullCheck;

/**
 * A {@link Processor} has variable number of {@link ForwardingThread}s waiting
 * for someone to put  work in the source queue. The work fetched from the queue
 * is then executed using a {@link ExecutorService} and a very simple
 * {@link Runnable} implementation named {@link WorkJob}.
  A fixed thread pool is used as {@link ExecutorService}. The number of threads
  to be used can be configured using the constructor.
 * 
 * What exactly is done with the work has to be implemented in the processWork()
 * hook method.
 * 
 * @author aw
 *
 *
 * @param <T> type of the work which can be found in the queue
 */
public abstract class Processor<T>  {
	
	
	/**
	 * Represents the source queue. This is where we can find some work.
	 */
	private Queue<T> mQueue;

	/**
	 * Contains references to the threads which are waiting for work to be put
	 * into the queue. These threads then forward it to the
	 * {@link ExecutorService} in form of a {@link Runnable} object.
	 * 
	 * @see WorkJob
	 * @see ForwardingThread
	 */
	private Thread mForwardThreads[];

	/**
	 * Represents the {@link ExecutorService} to process the work in the form of
	 * {@link Runnable} objects.
	 */
	private ExecutorService mExecutorService;

	/**
	 * Gives the number of worker threads in the thread pool.
	 */
	private int mWorkersCount;

	/**
	 * Gives the number of threads forwarding work from the queue to the
	 * thread pool.
	 */
	private int mForwardersCount;
	
	/**
	 * Creates a processor.
	 * 
	 * 
	 * @param queue source queue for work.
	 * @param forwardThreadsN number of {@link ForwardingThread}s to be used.
	 * @param workerThreadsN number of threads used in fixed thread pool
	 */
	public Processor(Queue<T> queue, int workers, int forwarders) {
		
		NullCheck.check(queue, "queue is null");
		if (workers < 1 || forwarders < 1) {
			throw new RuntimeException("must be initialized with > 0 threads");
		}
		mQueue = queue;
		mWorkersCount = workers;
		mForwardersCount = forwarders;
		mExecutorService = Executors.newFixedThreadPool(mWorkersCount);
		mForwardThreads = new Thread[mForwardersCount];
		
		for (int i = 0; i < mForwardThreads.length; i++) {
			mForwardThreads[i] = new ForwardingThread();
		}
	}
	
	
	public int getForwardersCount() {
		return mForwardersCount;
	}
	
	public int getWorkersCount() {
		return mWorkersCount;
	}
	
	
	/**
	 * Start the {@link Processor} by starting all {@link ForwardingThread}s.
	 */
	public void start() {
		for (int i = 0; i < mForwardThreads.length; i++) {
			mForwardThreads[i].start();
		}
	}
	
	/**
	 * Stop the {@link Processor}.
	 * 
	 * We interrupt all {@link ForwardingThread}s, which should make them stop,
	 * so we can join() them. Then we stop the {@link ExecutorService}.
	 */
	public void stop() {
		for (int i = 0; i < mForwardThreads.length; i++) {
			mForwardThreads[i].interrupt();
		}
		for (int i = 0; i < mForwardThreads.length; i++) {
			try {
				mForwardThreads[i].join();
			} catch (InterruptedException e) {
				// try again...
				i--;
			}
		}

		mExecutorService.shutdown();
		while (!mExecutorService.isShutdown()) {
			try {
				mExecutorService.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
			} catch (InterruptedException e) {
				// execute while() again, we wait until the executor service is
				// shut down.
			}
		}
	}
	
	
	
	/**
	 * Hook method to do actual work on the elements which are taken
	 * out of the queue. This method has to be reentrant because it may be
	 * executed by multiple threads at the same time.
	 * 
	 * @param e
	 */
	public abstract void processWork(T work);
	
	

	
	/**
	 * Simple {@link Runnable} implementation.
	 * 
	 * Call the hook method processWork() of the outer class.
	 * 
	 * @author aw
	 *
	 */
private class WorkJob implements Runnable {
		
		private T mWork;
		
		public WorkJob(T work) {
			mWork = work;
		}

		@Override
		public void run() {
			processWork(mWork);
		}
	}
	
	/**
	 * A simple thread which blocking waits to get some work out of the source
	 * queue. If there is new work to do a {@link WorkJob} object is created.
	 * This object is then executed by the {@link ExecutorService}.
	 * 
	 * If a interrupt is received at any point, the thread will return from
	 * the run() method and thereby stop running.
	 * 
	 * @author aw
	 */
	private class ForwardingThread extends Thread {
		
		// If the thread is blocked (mQueue.get()) and in the meanwhile
		// interrupted  a InterruptedException is thrown.
		// If the thread isn't blocked, the interrupt flag is set.
		// We break out of the while() when a InterruptException occurs,
		// we do not enter the while() anymore if the interrupt flag is set.

		public void run() {
			while (!this.isInterrupted()) {
				try {
					
					T work = mQueue.get();
					Runnable workJob = new WorkJob(work);
					mExecutorService.execute(workJob);
					
				} catch (InterruptedException e) {
					break;										/* BREAK */
				}
			}
		}
		
	}
}
