/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.communication.http
 * File:    ChannelThread.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.communication.http;

import java.io.IOException;
import java.io.InputStream;
import java.net.Socket;
import java.util.concurrent.Semaphore;

import org.apache.http.HttpEntity;
import org.apache.http.HttpEntityEnclosingRequest;
import org.apache.http.HttpException;
import org.apache.http.HttpRequest;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.HttpVersion;
import org.apache.http.entity.InputStreamEntity;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.DefaultHttpServerConnection;
import org.apache.http.message.BasicHeader;
import org.apache.http.message.BasicHttpResponse;
import org.apache.http.params.BasicHttpParams;
import org.apache.log4j.Logger;

import de.fhhannover.inform.iron.mapserver.communication.ChannelIdentifier;
import de.fhhannover.inform.iron.mapserver.communication.ClientIdentifier;
import de.fhhannover.inform.iron.mapserver.communication.bus.Queue;
import de.fhhannover.inform.iron.mapserver.communication.bus.messages.ClosedChannelEvent;
import de.fhhannover.inform.iron.mapserver.communication.bus.messages.Event;
import de.fhhannover.inform.iron.mapserver.communication.bus.messages.RequestChannelEvent;
import de.fhhannover.inform.iron.mapserver.logging.Logging;

/**
 * Accepts the http-request from client, and if authenticated, creates an event
 * and put it in the event-queue.
 * 
 * @author tr
 * 
 */
public class ChannelThread implements Runnable {

	private ChannelIdentifier mChannelIdentifier;
	private ChannelAuth mChannelAuth;
	private Socket mSocket;
	private Queue<Event> mQueue;	
	private boolean mIsDone;
	private DefaultHttpServerConnection mHttpConnection;
	private final Semaphore mSemaphore;
	private Logger mLogger;
	private int mCounter;

	public ChannelThread(Socket socket, ChannelAuth auth,
			ChannelIdentifier cid, Queue<Event> queue, int counter) {
		mSocket = socket;
		mChannelAuth = auth;
		mChannelIdentifier = cid;
		mQueue = queue;
		mCounter = counter;
		mSemaphore = new Semaphore(1, true);
		mLogger = Logging.getTheLogger();
		init();
	}

	private void init() {
		try {
			mHttpConnection = getDefaultHttpServerConnection(mSocket);
		} catch (IOException e) {
			mLogger.warn("ChannelThread.init: Cannot bind socket to DefaultHttpServerConnection");
		}
	}

	private DefaultHttpServerConnection getDefaultHttpServerConnection(Socket s)
			throws IOException {
		DefaultHttpServerConnection hsc = new DefaultHttpServerConnection();
		hsc.bind(s, new BasicHttpParams());
		return hsc;
	}

	public void setIsDone(boolean isDone) {
		mIsDone = isDone;
	}

	public boolean isDone() {
		return mIsDone;
	}

	public ChannelIdentifier getChannelIdentifier() {
		return mChannelIdentifier;
	}

	public ClientIdentifier getClientIdentifier() {
		return (mChannelAuth != null) ? mChannelAuth.getClientIdentifier() : null;
	}

	public void reply(InputStream reply) {
		try {
			int size = reply.available();
			BasicHeader header = new BasicHeader("Content-Length", "" + size);
			HttpResponse response = new BasicHttpResponse(HttpVersion.HTTP_1_1,
					HttpStatus.SC_OK, "");
			response.addHeader(header);
			HttpEntity respEntity = new InputStreamEntity(reply, size);
			response.setEntity(respEntity);
			mHttpConnection.sendResponseHeader(response);
			mHttpConnection.sendResponseEntity(response);
			respEntity.consumeContent();
			mSemaphore.release();
		} catch (HttpException e) {
			mLogger.warn("ChannelThread.reply: Sending response failed.");
			mSemaphore.release();
		} catch (IOException e) {
			mLogger.warn("ChannelThread.reply: Sending response failed.");
			mSemaphore.release();
		}
	}

	public void close() {
		try {
			mSocket.close();
		} catch (IOException e) {
			mLogger.warn("ChannelThread.close: Closing socket failed.");
		}	
		abort();
	}

	public void abort() {		
		mIsDone = true;		
		mChannelIdentifier = null;		
	}

	@Override
	public void run() {
		while (!isDone()) {
			try {
				mSemaphore.acquire();
				HttpEntity reqEntity = getHttpEntity();
				Event event = new RequestChannelEvent(getChannelIdentifier(),
						getClientIdentifier(), reqEntity.getContent());
				mQueue.put(event);
			} catch (HttpException e) {
				mLogger.debug("HttpException: " + e.getMessage());
			} catch (IOException e) {
				mLogger.debug("IOException: "+ e.getMessage());
				Event event = new ClosedChannelEvent(getChannelIdentifier());
				try {
					mQueue.put(event);
				} catch (InterruptedException e1) {
					mLogger.debug("mQueue.put interrupted.");
				}
			} catch (AuthException e) {
				mLogger.debug("Authentification failed: " + e.getMessage());
				sendUnauthorizedResponse();
				mSemaphore.release();
			} catch (InterruptedException e) {
				// TODO The queue throws an InterruptedException. Handle it!
				mLogger.debug("Interrupted.");
				close();
			}
		}
	}
	
	private void sendUnauthorizedResponse() {		
		String html = "<html><head><title>401 Unauthorized</title></head></html>";
		HttpResponse denied = new BasicHttpResponse(HttpVersion.HTTP_1_1,
				HttpStatus.SC_UNAUTHORIZED, "Unauthorized");
		BasicHeader header = new BasicHeader("WWW-Authenticate", "Basic realm=\""
					+ mSocket.getInetAddress().getHostAddress()+"\"");

		denied.addHeader(header);
		header = new BasicHeader("Content-length", "" + html.length());
		denied.addHeader(header);				
		try {
			StringEntity entity = new StringEntity(html);
			denied.setEntity(entity);			
			mHttpConnection.sendResponseHeader(denied);
			mHttpConnection.sendResponseEntity(denied);			
			entity.consumeContent();
		} catch (HttpException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	}

	private HttpEntity getHttpEntity() throws HttpException, IOException,
			AuthException {
		HttpRequest req = mHttpConnection.receiveRequestHeader();
		HttpEntity reqEntity = null;
		if (req instanceof HttpEntityEnclosingRequest) {
			mHttpConnection
					.receiveRequestEntity((HttpEntityEnclosingRequest) req);
			reqEntity = ((HttpEntityEnclosingRequest) req).getEntity();
		}

		try {
			mChannelAuth.authenticate(req);
		} catch (AuthException e) {
			if (reqEntity != null) {
				reqEntity.consumeContent();
			}
			throw e;
		}
		
		// here we are authenticated
		
		if (reqEntity == null) {
			throw new HttpException();
		}
		return reqEntity;
	}
}
