/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel
 * File:    MetaController.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel;

import java.util.List;
import java.util.Map;


import de.fhhannover.inform.iron.mapserver.datamodel.graph.Identifier;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Link;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.MetadataContainer;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.MetaState;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;

/**
 * MetaController
 * 
 * @author vp
 * @version 0.1
 */

/*
 * created: 10.02.10
 * changes:
 * 	11.02.10 vp - the visibility of the classes and methods is changed
 *  30.04.10 aw - make methods non static, remove singletons!
 *  11.05.10 aw - Subscriptions:
 *  				When deleting metadata don't remove it, just mark it as
 *                  removed. This way we can update subscriptions later on.
 *
 */
class MetaController {

	/**
	 * Logger instance
	 *
	private static Logger logger;
	static {
		logger = Logging.getTheLogger();
	}
	*/

	private IdentifierRep identifierRep;
	private LinkRep linkRep;
	
	public MetaController(IdentifierRep iRep, LinkRep lRep) {
		identifierRep	= iRep;
		linkRep			= lRep;
	}
	 
	/**
	 * Helper method to mark {@link Metadata} objects as deleted or remove
	 * them from the graph.
	 * 
	 * If a {@link Metadata} object has been in the graph before the currently
	 * running publish request, i.e. m.isInGraph() == TRUE && m.isNew() == FALSE,
	 * the object is marked as "deleted". However it is not removed from the graph.
	 * If the metadata wasn't in the graph before, e.g. it was just published
	 * in the same request, it can be safely removed from the graph. publish update
	 * and publish delete in one publish request can be canceled out.
	 * 
	 * {@link Metadata} objects that are marked as removed are added into the changes
	 * map. If {@ Metadata} objects are dropped as it was new this metadata is removed
	 * from the changes list, as they don't exist in the graph anymore and it
	 * won't make sense to process them later in commitChanges()
	 * 
	 * 
	 * @param reqpublisher The publisher that made the delete request.
	 * @param toremove A list of metadata references which will be marked deleted
	 * @param changes A map where "changed" Metadata objects are stored
	 * 
	 */
	void deleteMetadata(Publisher reqpublisher, List<Metadata> toremove,
			Map<Integer, Metadata> changes) {
		Publisher metapublisher;
		for (Metadata m : toremove) {
			if (m.isNew()) {
				changes.remove(m.hashCode());
				m.getContainerReference().removeMetadata(m);
				metapublisher = m.getPublisherReference();
				metapublisher.removeMetadata(m);
			} else {
				m.setState(MetaState.DELETED);
				changes.put(m.hashCode(), m);
			}
		}
	}
	
	void removeReferencesTo(Metadata m) {
		if (m != null) {
			m.getContainerReference().removeMetadata(m);
			m.getPublisherReference().removeMetadata(m);
			removeEmptyMetadataContainer(m.getContainerReference());
		}
		
	}
	
	/**
	 * Remove Identifier without metadata and links.
	 * Remove Links without metadata.
	 * 
	 * @param metadatacontainer
	 */
	private void removeEmptyMetadataContainer (MetadataContainer metadatacontainer) {
		
		if (!metadatacontainer.getSubscriptions().isEmpty()) {
			return;
		}
		
		if (metadatacontainer.getMetadata().isEmpty()) {
			if (metadatacontainer instanceof Identifier) {
				Identifier id = (Identifier) metadatacontainer;
				if (id.getLinks().isEmpty()) {
					identifierRep.removeIdentifier((Identifier)metadatacontainer);
				}
			} 
			
			if (metadatacontainer instanceof Link) {
				Link link = (Link) metadatacontainer;
				linkRep.removeLink(link);
				removeEmptyMetadataContainer(link.getIdent1());
				removeEmptyMetadataContainer(link.getIdent2());
			}
		}
	}
}