/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel
 * File:    PublishService.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import de.fhhannover.inform.iron.mapserver.datamodel.IdentifierRep;
import de.fhhannover.inform.iron.mapserver.datamodel.LinkRep;
import de.fhhannover.inform.iron.mapserver.datamodel.PublisherRep;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.DoRollbackException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.InvalidMetadataException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.NoSuchPublisherException;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Identifier;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.MetadataContainer;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.MetaState;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;
import de.fhhannover.inform.iron.mapserver.logging.Logging;
import de.fhhannover.inform.iron.mapserver.messages.PublishDelete;
import de.fhhannover.inform.iron.mapserver.messages.PublishNotify;
import de.fhhannover.inform.iron.mapserver.messages.PublishUpdate;
import de.fhhannover.inform.iron.mapserver.messages.PublishRequest;
import de.fhhannover.inform.iron.mapserver.messages.SubPublishRequest;

/**
 * Publish or delete metadata.
 *
 *
 * TODO: Check what should be done with replaced singleValue metadata.
 *       Currently they are silently removed. Do they have to appear in
 *       poll results of MAPCs?
 * 
 * @author awe, vp
 * @version 0.1
 * created: 19.11.09
 */

/*
 * changes:
 *  22.11.09 vp - processPublishUpdate() and processPublishDelete() parameter changed and implemented, 
 * 		publish(), getMetadataContainer() implemented
 *  23.11.09 aw - changed instanceof checks.
 *  05.12.09 aw - changed position of casts for process***Requests, get the
 *  	publisher of this publishRequest in publish(),
 *  	added Singleton functions.
 *  10.12.09 aw - syserrs
 *  17.12.09 vp - set realation between publisher and metadata.
 *  22.12.09 vp - changed processPublishDelete 
 *  	remove identifier without links and metadata and remove links without metadata
 *  	from repositories
 *  22.12.09 vp - removeEmptyMetadata() created
 *  04.02.10 aw - logging with log4j, setting timestamp
 *  05.02.10 aw - added calls to ResponseFactory
 *  08.02.10 aw - removing of metadata simplified, this class now removes
 *                metadata from the container as well as from the publisher
 *                no dangling calls through the metadatacontainer to the publisher
 *  10.02.10 aw - processPublishUpdate() now sets references instead of
 *                the Metadatacontainer method, process* methods now return
 *                boolean values instead of responses., some comments, some
 *                cleanup
 *  10.02.10 vp - removing of meta from original publisher is in MetaController.java 
 *  11.02.10 vp - the visibility of the classes and methods is changed
 *  10.04.10 aw - setting lifetime on metadata is done at higher levels
 *  30.04.10 aw - remove singletons!, remove responsefactory, make methods void
 *                in case of an error they throw an exception
 *  30.09.10 aw - handle Rollbacks for section 3.3.3 of the specification
 */
class PublishService {
	
	/**
	 * a static logger instance
	 */
	private static Logger logger;
	static {
		logger = Logging.getTheLogger();
	}
	
	private PublisherRep publisherRep;
	private IdentifierRep identifierRep;
	private MetaController metaController;
	private LinkRep linkRep;
	private SubscriptionService subService;
	
	PublishService(PublisherRep pRep, IdentifierRep iRep, LinkRep lRep,
			MetaController mCtl, SubscriptionService subServ) {
		publisherRep	= pRep;
		identifierRep	= iRep;
		linkRep			= lRep;
		metaController	= mCtl;
		subService		= subServ;
	}
	
	/**
	 * Go through the whole PublishRequest and dispatch the elements to
	 * the corresponding process* methods
	 *
	 * @param request
	 * @return PublishReceived or PublishError
	 * @throws NoSuchPublisherException 
	 * @throws NotSupportedException 
	 * @throws InvalidMetadataException 
	 * @throws ResponseCreationException 
	 */
	void publish(PublishRequest request) throws NoSuchPublisherException, InvalidMetadataException {
		logger.trace("PublishRequest from sessionid=" + request.getSessionId());
		Iterator<SubPublishRequest> it = request.getSubPublishRequestList().iterator();
		
		Publisher publisher = 
			publisherRep.getPublisherBySessionId(request.getSessionId());
		
		logger.trace("Traversing n=" + request.getSubPublishRequestList().size() + "" +
				" PublishRequests");
		
		Map<Integer, Metadata> changes = new HashMap<Integer, Metadata>();
		
		try {
			while(it.hasNext()) {
				SubPublishRequest subpublishrequest = it.next();
			
				
				if (subpublishrequest instanceof PublishNotify) {
					
					processPublishNotify(publisher, (PublishNotify) subpublishrequest, changes);
					
				} else if (subpublishrequest instanceof PublishUpdate) {
					
					processPublishUpdate(publisher, (PublishUpdate)subpublishrequest, changes);
						
				} else if (subpublishrequest instanceof PublishDelete) {
					
					processPublishDelete(publisher, (PublishDelete)subpublishrequest, changes);
				}
			}
		
			// set timestamp and publisher id _now_;
			String timeNow = getISO8601DateTime(new Date());
			for (Metadata m : changes.values()) {
				m.setTimestamp(timeNow);
				m.setPublisherId(publisher.getPublisherId());
			}
			
			identifierRep.setTimestamp(System.currentTimeMillis());
			subService.updateSubscriptions(changes);
		} catch (DoRollbackException e) {
			logger.debug("Doing Rollback!");
			subService.rollbackChanges(changes);
			throw new InvalidMetadataException(e.getMessage());
		}
	}
	
	
	/**
	 * Publish metadata to identifier or link.
	 * 
	 * - get a MetadataContainer with help of getMetadataContainer()
	 *   (whether link or identifier does not matter)
	 * 
	 * - go through the list of metadata, modify every metadata object
	 *   to include the publisherId, a timestamp and set the lifetime
	 *   thereafter call addMetadata() on the target container.
	 *   a copy of the metadata will be created an returned.
	 *   
	 * - add the new metadata object to the publisher who published it
	 * 
	 *
	 * @param request
	 * @throws DoRollbackException 
	 */
	private void processPublishUpdate(Publisher pub, PublishUpdate request,
			Map<Integer, Metadata> changed) throws DoRollbackException {
		
		List<Metadata> mlist = request.getMetadataList();
		
	
		// set the state of all metadata objects, if it's not NOTIFY (set by
		// processPublishNotify(), then it must be NEW
		for (Metadata m : mlist) {
			if (!m.isNotify()) {
				m.setState(MetaState.NEW);
			}
		}
		
		MetadataContainer metadatacontainer = 
			getMetadataContainer(request.getIdent1(),request.getIdent2());
		
		logger.trace(pub + " is adding n=" + mlist.size() + " Metadata to " +
				metadatacontainer + " with lifetime=" + request.getLifeTime()); 
		

		for (Metadata m : mlist) {
			List<Metadata> removed = new LinkedList<Metadata>();
			
			Metadata added = metadatacontainer.addMetadata(m, removed);
		
			if (added != null) {
				changed.put(added.hashCode(), added);
				added.setContainerReference(metadatacontainer);
				added.setPublisherReference(pub);
				pub.addMetadata(added);
				
				// snaity check
				if (added.isNotify()) {
					if (removed.size() > 0) {
						logger.error("UNEXPECTED: notify metadata triggered removal" +
								" of metadata in the graph");
					}
				}
			
				// check whether new metadata was replaced, if yes remove it from
				// changes
				for (Metadata rmmd : removed) {
					if (rmmd.isNew()) {
						// sanity check
						if (changed.remove(rmmd.hashCode()) != null) {
							throw new RuntimeException("NEW Metadata not in changes." +
									"Programming Error?");
						}
					} else if (rmmd.isDeleted()){
						// leave deleted in the changes and "in the graph"
						// FIXME: is this ok?
					} else {
						// silently drop removed metadata
						metaController.removeReferencesTo(rmmd);
					}
				}
			} else {
				logger.error("Metadata adding failed!");
			}
		}
	}
	 
	/**
	 * Delete metadata from link or identifier
	 * 
	 * @param reqPublisher the publisher responsible for this request
	 * @param request the delete request
	 * @param changed 
	 * @return
	 */
	private void processPublishDelete(Publisher reqPublisher, PublishDelete request,
			Map<Integer, Metadata> changed) {
		
		List<Metadata> toremove;
		MetadataContainer metadatacontainer;
		
		metadatacontainer = getMetadataContainer(request.getIdent1(),
				request.getIdent2());
		
		// request metadata to be removed
		toremove = metadatacontainer.getMetadata(request.getFilter());
		
		
		logger.trace("Deleting n=" + toremove.size() + " Metadata objects from "
				+ metadatacontainer);
	
		metaController.deleteMetadata(reqPublisher, toremove, changed);		
	}
	
	
	
	
	/**
	 * Get the list of {@link Metadata} objects, set their state to
	 * NOTIFY and forward the request to the processPublishUpdate
	 * method. There are some hacks over there that this process
	 * should work...
	 * 
	 * @param pub
	 * @param request
	 */
	private void processPublishNotify(Publisher pub, PublishNotify request,
			Map<Integer, Metadata> changed) {
		
		List<Metadata> mlist = request.getMetadataList();
		
		
		for (Metadata m : mlist) {
			m.setState(MetaState.NOTIFY);
		}
		try {
			processPublishUpdate(pub, request, changed);
		} catch (DoRollbackException e) {
			// nah, that should not occur...
			logger.error("UNEXPECTED: Rollback triggered while using notify metadata");
			e.printStackTrace();
		}
	}
	
	

	/**
	 * Check data type of SubPublishRequest and return MetadataContainer
	 * 
	 * @param ident1
	 * @param ident2
	 * @return MetadataContainer
	 */
	private MetadataContainer getMetadataContainer(Identifier ident1, Identifier ident2) {
		MetadataContainer metadatacontainer = null;
		if (ident1 != null && ident2 != null) {
			metadatacontainer = 
				linkRep.getOrCreateLink(ident1, ident2);
		} else if (ident1 != null || ident2 != null) {
			
			if (ident1 != null) {
				metadatacontainer = 
					identifierRep.getOrCreateIdentifier(ident1);
			} else {
				metadatacontainer = 
					identifierRep.getOrCreateIdentifier(ident2);
			}
		} else {
			metadatacontainer = null;
		}
		return metadatacontainer;
	}
	
	/**
	 * Thanks Ingo!
	 * @param dt
	 * @return
	 */
	private String getISO8601DateTime(Date dt){
		//FIXME hack in order to get good timestamps for xsd:dateTime
        SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
        SimpleDateFormat tzFormatter = new SimpleDateFormat("Z");
        String timezone = tzFormatter.format(dt);
        return formatter.format(dt) + timezone.substring(0, 3) + ":"
                + timezone.substring(3);
	}
}
 

