/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.utils
 * File:    JaxbRequestTransformer.java
 *
 * Copyright (C) 2010 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.utils;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.bind.Binder;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.trustedcomputinggroup._2010.ifmap._2.DeleteSearchRequestType;
import org.trustedcomputinggroup._2010.ifmap._2.DeleteType;
import org.trustedcomputinggroup._2010.ifmap._2.DumpRequestType;
import org.trustedcomputinggroup._2010.ifmap._2.EndSessionType;
import org.trustedcomputinggroup._2010.ifmap._2.MetadataListType;
import org.trustedcomputinggroup._2010.ifmap._2.NewSessionRequestType;
import org.trustedcomputinggroup._2010.ifmap._2.NotifyType;
import org.trustedcomputinggroup._2010.ifmap._2.PollRequestType;
import org.trustedcomputinggroup._2010.ifmap._2.PublishRequestType;
import org.trustedcomputinggroup._2010.ifmap._2.PurgePublisherRequestType;
import org.trustedcomputinggroup._2010.ifmap._2.RenewSessionType;
import org.trustedcomputinggroup._2010.ifmap._2.SearchRequestType;
import org.trustedcomputinggroup._2010.ifmap._2.SearchType;
import org.trustedcomputinggroup._2010.ifmap._2.SubscribeRequestType;
import org.trustedcomputinggroup._2010.ifmap._2.UpdateType;
import org.w3._2003._05.soap_envelope.Body;
import org.w3._2003._05.soap_envelope.Envelope;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import de.fhhannover.inform.iron.mapserver.datamodel.Filter;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.IdentifierConstructionException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.MetadataConstructionException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.RequestCreationException;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Identifier;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.MetaLifeTimeType;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.W3cXmlMetadata;
import de.fhhannover.inform.iron.mapserver.messages.DumpRequest;
import de.fhhannover.inform.iron.mapserver.messages.EndSessionRequest;
import de.fhhannover.inform.iron.mapserver.messages.NewSessionRequest;
import de.fhhannover.inform.iron.mapserver.messages.PollRequest;
import de.fhhannover.inform.iron.mapserver.messages.PublishRequest;
import de.fhhannover.inform.iron.mapserver.messages.PurgePublisherRequest;
import de.fhhannover.inform.iron.mapserver.messages.RenewSessionRequest;
import de.fhhannover.inform.iron.mapserver.messages.Request;
import de.fhhannover.inform.iron.mapserver.messages.RequestFactory;
import de.fhhannover.inform.iron.mapserver.messages.SearchRequest;
import de.fhhannover.inform.iron.mapserver.messages.SubPublishRequest;
import de.fhhannover.inform.iron.mapserver.messages.SubSubscribeRequest;
import de.fhhannover.inform.iron.mapserver.messages.SubscribeDelete;
import de.fhhannover.inform.iron.mapserver.messages.SubscribeRequest;

/**
 * Implementation of the {@link RequestTransformer} interface using JAXB
 * for the XML binding.
 * 
 * Functionality originates from the old MessageTransformer class. It was
 * a mess back then and it's still a mess, but we now hide it behind an
 * interface.
 * 
 * @author aw
 */
class JaxbRequestTransformer implements RequestTransformer {
	
	private JAXBContext mJaxbCtx; 
	private static DocumentBuilderFactory mDocumentBuilderFactory;
	private ThreadLocal<Binder<Node>> mBinder;
	
	// FIXME: grml grml grml ...singletons... grml grml grml
	private FilterTransformer fTrans = FilterTransformer.getInstance();
	private JaxbIdentifierExtractor iExtractor = JaxbIdentifierExtractor.getInstance();
	
	private RequestFactory requestFactory = RequestFactory.getInstance();
	
	/**
	 * Constructs
	 * 
	 * @param serverConf
	 */
	JaxbRequestTransformer() {
		
		try {
			mJaxbCtx = JAXBContext.newInstance(Envelope.class);
			mBinder = new ThreadLocal<Binder<Node>>() {
				@Override
				protected Binder<Node> initialValue() {
					return mJaxbCtx.createBinder();
				}
			};
		} catch (JAXBException e) {
			// go crazy if we can't initialize the JAXB context or something
			// JAXB related
			throw new RuntimeException(e);
		}
	}
	
	static {
		mDocumentBuilderFactory = DocumentBuilderFactory.newInstance();
		mDocumentBuilderFactory.setNamespaceAware(true);
		// done by JAXB?
		mDocumentBuilderFactory.setValidating(false);
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public Request transform(InputStream requestContent) throws RequestCreationException,
																IdentifierConstructionException,
																MetadataConstructionException {
		DocumentBuilder documentBuilder = null;
		try {
			documentBuilder = mDocumentBuilderFactory.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			// TODO Go Crazy? If we happen to end up here, game's over anyway...
			e.printStackTrace();
			throw new RequestCreationException(e.getMessage());
		}
		
		Document requestDocument = null;
		
		try {
			requestDocument = documentBuilder.parse(requestContent);
		} catch (SAXException e) {
			// TODO Go Crazy? If we happen to end up here, game's over anyway...
			e.printStackTrace();
			throw new RequestCreationException(e.getMessage());
		} catch (IOException e) {
			// TODO Go Crazy? If we happen to end up here, game's over anyway...
			e.printStackTrace();
			throw new RequestCreationException(e.getMessage());
		}
		
		JAXBElement<Envelope> reqEnvJaxb = null;
		try {
			reqEnvJaxb = (JAXBElement<Envelope>) mBinder.get().unmarshal(requestDocument);
		} catch (JAXBException e) {
			//e.printStackTrace();
			throw new RequestCreationException(e.getMessage());
		}
		
		
		Envelope reqEnv = reqEnvJaxb.getValue();

		Body soapBody = reqEnv.getBody();
		
		if (soapBody == null) {
			throw new RequestCreationException("no body found");
		}
	
		
		if (soapBody.getDump() != null) {
			
			return transformDumpRequest(soapBody.getDump());
			
		} else if (soapBody.getEndSession() != null) {
			
			return transformEndSessionRequest(soapBody.getEndSession());
			
		} else if (soapBody.getNewSession() != null) {
			
			return transformNewSessionRequest(soapBody.getNewSession());
			
		} else if (soapBody.getPoll() != null) {
			
			return transformPollRequest(soapBody.getPoll());
			
		} else if (soapBody.getPublish() != null) {
			
			return transformPublishRequest(soapBody.getPublish());
			
		} else if (soapBody.getPurgePublisher() != null) {
			
			return transformPurgePublisherRequest(soapBody.getPurgePublisher());
			
		} else if (soapBody.getRenewSession() != null) {
			
			return transformRenewSessionRequest(soapBody.getRenewSession());
			
		} else if (soapBody.getSearch() != null) {
			
			return transformSearchRequest(soapBody.getSearch());
			
		} else if (soapBody.getSubscribe() != null) {
			
			return transformSubscribeRequest(soapBody.getSubscribe());
			
		} else {
			
			throw new RequestCreationException("No IF-MAP Operation found");
			
		}
	}
	
	private Map<String, String> extractNsMap(Node node) {
		Map<String, String> ret = new HashMap<String, String>();
		extractNsMap(node, ret);
		return ret;
	}
	
	private void extractNsMap(Node w3cNode, Map<String, String> nsMap) {

		if (w3cNode == null) {
			return;
		}
	
		if (w3cNode.getNodeType() == Node.DOCUMENT_NODE) {
			return;
		}
		
		NamedNodeMap nnm = w3cNode.getAttributes();
		
		for (int i = 0; i < nnm.getLength(); i++) {
			Node node = nnm.item(i);
			if (node.getNodeType() == Node.ATTRIBUTE_NODE) {
				String nodeName = node.getNodeName();
				String nodeValue = node.getNodeValue();
				if (nodeName.startsWith("xmlns:")) {
					String prefix = nodeName.substring(6); // magic number xmlns:6
					// don't overwrite existing mappings, the inner ones are those
					// that we want to have!
					if (!nsMap.containsKey(prefix)) {
						nsMap.put(nodeName.substring(6), nodeValue);
					}
				}
			}
		}
		
		extractNsMap(w3cNode.getParentNode(), nsMap);
	}

	
	private DumpRequest transformDumpRequest(DumpRequestType dump)
			throws RequestCreationException {
		String sessionId = dump.getSessionId();
		String identifier = dump.getIdentifier();
		return requestFactory.createDumpRequest(sessionId, identifier);
	}

	
	private EndSessionRequest transformEndSessionRequest(EndSessionType endSession)
			throws RequestCreationException {
		String sessionId = endSession.getSessionId();
		return requestFactory.createEndSessionRequest(sessionId);
	}

	
	private NewSessionRequest transformNewSessionRequest(NewSessionRequestType newSession)
			throws RequestCreationException {
		long mprsParameter = NewSessionRequest.MAX_POLL_RESULT_NOT_GIVEN;
		BigInteger mprsReceived = newSession.getMaxPollResultSize();
		
		if (mprsReceived != null) {
			mprsParameter = mprsReceived.longValue();
		}
			
		return requestFactory.createNewSessionRequest(mprsParameter);
	}

	private PollRequest transformPollRequest(PollRequestType poll)
			throws RequestCreationException {
		String sessionId = poll.getSessionId();
		return requestFactory.createPollRequest(sessionId);
	}

	
	private PurgePublisherRequest transformPurgePublisherRequest( PurgePublisherRequestType purgePublisher)
			throws RequestCreationException {
		String sessionId = purgePublisher.getSessionId();
		String publisherId = purgePublisher.getIfmapPublisherId();
		return requestFactory.createPurgePublisherRequest(sessionId, publisherId);
	}
	

	private RenewSessionRequest transformRenewSessionRequest(RenewSessionType renewSession)
			throws RequestCreationException {
		String sessionId = renewSession.getSessionId();
		return requestFactory.createRenewSessionRequest(sessionId);
	}
	
	/**
	 * 
	 * 
	 * @param search
	 * @return
	 * @throws RequestCreationException
	 * @throws IdentifierConstructionException
	 */
	private SearchRequest transformSearchRequest(SearchType search, String sessionID)
			  throws RequestCreationException, IdentifierConstructionException {
		
		Node w3cSearch = null;
		SearchRequest searchRequest = null;
		if (search != null) {
			w3cSearch = mBinder.get().getXMLNode(search);
			Map<String, String> nsMap = extractNsMap(w3cSearch);
			Identifier ident = iExtractor.extractFromSearch((SearchType)search);
			Long maxDepth = search.getMaxDepth();
			Long maxSize = search.getMaxSize();
			// if no depth is given, process it with zero depth
			int maxDepthi = (maxDepth == null) ? 0 : maxDepth.intValue();
			int maxSizei  = (maxSize == null) ? SearchRequest.MAX_RESULT_SIZE_NOT_GIVEN : maxSize.intValue();
			Filter matchLinks = fTrans.transformFilter(search.getMatchLinks(), nsMap);
			Filter resultFilter = fTrans.transformFilter(search.getResultFilter(), nsMap);
			if (search instanceof SearchRequestType) {
				sessionID = ((SearchRequestType)search).getSessionId();
			}
			
			searchRequest = requestFactory.createSearchRequest(sessionID,
					maxDepthi, maxSizei, search.getTerminalIdentifierType(), ident,
					matchLinks, resultFilter);
		}
		return searchRequest;
	}
	



	private SearchRequest transformSearchRequest(SearchType search) throws 
							RequestCreationException, IdentifierConstructionException {
		return transformSearchRequest(search, null);
	}	
	
	
	private PublishRequest transformPublishRequest(PublishRequestType publish)
					throws IdentifierConstructionException, RequestCreationException,
													MetadataConstructionException{
		PublishRequest publishRequest = null;
		List<Object> list = null;
		List<SubPublishRequest> reqlist = null;
		
		if (publish != null){
			String sid = publish.getSessionId();

			list = publish.getUpdateOrNotifyOrDelete();
			reqlist = transformToSubPublishRequestList(list);
			publishRequest = requestFactory.createPublishRequest(sid, reqlist);
		}
		return publishRequest;
	}

	
	private SubscribeRequest transformSubscribeRequest (SubscribeRequestType srt) 
						throws RequestCreationException, IdentifierConstructionException {
		
		SubscribeRequest sr = null;
		List<Object> list = null;
		List<SubSubscribeRequest> sublist = null;
		
		if (srt != null) {
			String sid = srt.getSessionId();
			list = srt.getUpdateOrDelete();
			sublist = transformToSubSubscribeList(list, srt.getSessionId());
			sr = requestFactory.createSubscribeRequest(sid, sublist);
		}
		
		return sr;
	}
	
	
	private List<SubSubscribeRequest> transformToSubSubscribeList(
			List<Object> list, String sessionID) 
			throws RequestCreationException, IdentifierConstructionException {
		
		ArrayList<SubSubscribeRequest> ssr = new ArrayList<SubSubscribeRequest>();
		for (int i = 0; i < list.size(); i++) {
			if (list.get(i) instanceof DeleteSearchRequestType) {
				ssr.add(new SubscribeDelete(
						((DeleteSearchRequestType) list.get(i)).getName()));
			} else if (list.get(i) instanceof SubscribeRequestType.Update) {
				SearchRequest sr = transformSearchRequest
							(((SubscribeRequestType.Update)list.get(i)), sessionID);
				
				ssr.add(requestFactory.createSubscribeUpdate(
						((SubscribeRequestType.Update)list.get(i)).getName(), sr));
			}
		}
		
		return ssr;
	}
	
	

	/**
	 * Wrapper to transform a list of objects which can be of
	 * type UpdateType, DeleteType and NotifyType into
	 * 
	 * 
	 * @param list
	 * @return
	 * @throws IdentifierConstructionException
	 * @throws RequestCreationException
	 * @throws MetadataConstructionException 
	 */
	private List<SubPublishRequest> transformToSubPublishRequestList(List<Object> list)
				throws IdentifierConstructionException, RequestCreationException,
													MetadataConstructionException{
		List<SubPublishRequest> retList = new ArrayList<SubPublishRequest>();
		
		for (Object o : list)
			retList.add(transformToSubPublishRequest(o));
			
		return retList;
	}
	
	

	private SubPublishRequest transformToSubPublishRequest(Object o)
			throws RequestCreationException, IdentifierConstructionException,
												MetadataConstructionException {
		SubPublishRequest ret = null;
		if (o != null) {
			if (o instanceof UpdateType) {
				ret = transformUpdateType((UpdateType)o);
			} else if (o instanceof DeleteType) {
				ret = transformDeleteType((DeleteType)o);
			} else if (o instanceof NotifyType) {
				ret = transformNotifyType((NotifyType)o);
			}
		}
		return ret;
	}

	private SubPublishRequest transformNotifyType(NotifyType value) throws 
									MetadataConstructionException, IdentifierConstructionException, RequestCreationException {
		SubPublishRequest ret = null;
		if (value != null) {
			MetaLifeTimeType lt;
			String ltstr = value.getLifetime();
			if (ltstr == null) {
			}
			if (ltstr.equals(MetaLifeTimeType.session.toString())) {
				lt = MetaLifeTimeType.session;
			} else {
				lt = MetaLifeTimeType.forever;
			}
			List<Metadata> mlist = transformMetadata(value.getMetadata(), lt);
			Identifier idents[] = iExtractor.extractFromNotify(value);
			ret = requestFactory.createPublishNotifyRequest(idents[0], idents[1],
					mlist, lt);
		}
		
		return ret;
	}


	private SubPublishRequest transformDeleteType(DeleteType value)
				throws IdentifierConstructionException, RequestCreationException {
		SubPublishRequest ret = null;
		Node w3cNode = null;
		if (value != null) {
			w3cNode = mBinder.get().getXMLNode(value);
			Identifier idents[] = iExtractor.extractFromDelete(value);
			Map<String, String> nsMap = extractNsMap(w3cNode);
			
			
// FIXME Use new transformFilter()               !!!
			Filter f = fTrans.transformFilter(value.getFilter(), nsMap);
			ret = requestFactory.createPublishDeleteRequest(idents[0], idents[1], f);
		}
		return ret;
	}
	
	

	private SubPublishRequest transformUpdateType(UpdateType value)
				throws IdentifierConstructionException, RequestCreationException,
												MetadataConstructionException {
		
		SubPublishRequest ret = null;
		if (value != null) {
			MetaLifeTimeType lt;
			String ltstr = value.getLifetime();
			if (ltstr == null) {
				// FIXME: WTF???
			}
			if (ltstr.equals(MetaLifeTimeType.session.toString())) {
				lt = MetaLifeTimeType.session;
			} else {
				lt = MetaLifeTimeType.forever;
			}
			if (value.getMetadata() == null) {
				throw new MetadataConstructionException("No Metadata in update element found.");
				
			}
			List<Metadata> mlist = transformMetadata(value.getMetadata(), lt);
			Identifier idents[] = iExtractor.extractFromUpdate(value);
			ret = requestFactory.createPublishUpdateRequest(idents[0], idents[1], mlist, lt);
		}
		
		return ret;
	}
	

	
	/**
	 * Method to transform the autogenerated MetadataList into a
	 * List of datamodel Metadata. 
	 * 
	 * This method transforms to the XMLMetadata implementation
	 * using DOMBuilder from JDOM to create a JDOM Document from
	 * the Document objects in the MetadataListType.
	 * 
	 * 
	 * @param metaDataListType
	 * @return List of Metadata that the datamodel can work with
	 * @throws MetadataConstructionException
	 */
	List<Metadata> transformMetadata(MetadataListType metaDataListType,
			MetaLifeTimeType lifetime) throws MetadataConstructionException {
		List<Metadata> lmeta = new ArrayList<Metadata>();
		List<Object> lo = metaDataListType.getAny();
		if (lo == null || lo.size() == 0) {
			throw new MetadataConstructionException("No metadata in metadata list found");
		}
		for (Object o : lo) {
			Metadata metadata;
			try {
				// i have no idea why this works...
				metadata = new W3cXmlMetadata((Node)o, lifetime);
			} catch (Exception e) {
				// throw up anything as a MetaDataConstructionException
				throw new MetadataConstructionException(e.getMessage());
			}
			lmeta.add(metadata);
		}
		return lmeta;
	}
	
	


}
